import unittest

import time
from redis import Redis
from multiprocessing import Process

from concrete.util.redis_io import (
    read_communication_from_redis_key,
    RedisCommunicationReader,
    write_communication_to_redis_key,
    RedisCommunicationWriter
)
from concrete.util.mem_io import (
    read_communication_from_buffer,
    write_communication_to_buffer
)
from concrete.util.simple_comm import create_comm
from redis_server import RedisServer


def _add_comm_to_list(sleep, port, comm_id, key):
    time.sleep(sleep)
    redis_db = Redis(port=port)
    comm = create_comm(comm_id)
    buf = write_communication_to_buffer(comm)
    redis_db.lpush(key, buf)


class TestReadCommunicationFromRedisKey(unittest.TestCase):

    def test_read_against_file_contents(self):
        filename = u'tests/testdata/simple_1.concrete'
        key = 'comm'
        with open(filename, 'rb') as f:
            buf = f.read()
            with RedisServer(loglevel='warning') as server:
                redis_db = Redis(port=server.port)
                redis_db.set(key, buf)
                comm = read_communication_from_redis_key(redis_db, key)

            self.assertTrue(hasattr(comm, 'sentenceForUUID'))
            self.assertEquals('one', comm.id)

    def test_read_against_file_contents_no_add_references(self):
        filename = u'tests/testdata/simple_1.concrete'
        key = 'comm'
        with open(filename, 'rb') as f:
            buf = f.read()
            with RedisServer(loglevel='warning') as server:
                redis_db = Redis(port=server.port)
                redis_db.set(key, buf)
                comm = read_communication_from_redis_key(
                    redis_db, key, add_references=False
                )

            self.assertFalse(hasattr(comm, 'sentenceForUUID'))
            self.assertEquals('one', comm.id)


class TestWriteCommunicationToRedisKey(unittest.TestCase):

    def test_write_against_file_contents(self):
        filename = u'tests/testdata/simple_1.concrete'
        key = 'comm'
        with open(filename, 'rb') as f:
            f_buf = f.read()
            comm = read_communication_from_buffer(f_buf)
            with RedisServer(loglevel='warning') as server:
                redis_db = Redis(port=server.port)
                write_communication_to_redis_key(redis_db, key, comm)
                self.assertEquals(f_buf, redis_db.get(key))

    def test_read_write_fixed_point(self):
        key = 'comm'
        comm = create_comm('comm-1')
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            buf_1 = write_communication_to_redis_key(redis_db, key, comm)
            buf_2 = write_communication_to_redis_key(
                redis_db, key,
                read_communication_from_redis_key(redis_db, key)
            )
            self.assertEquals(buf_1, buf_2)


class TestRedisCommunicationReader(unittest.TestCase):

    def setUp(self):
        self.comm1 = create_comm('comm-1')
        self.comm2 = create_comm('comm-2')
        self.comm3 = create_comm('comm-3')

    def test_set(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.sadd(key, write_communication_to_buffer(self.comm1))
            redis_db.sadd(key, write_communication_to_buffer(self.comm2))
            redis_db.sadd(key, write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key, key_type='set')
            comms = [c for c in reader]
            ids = [c.id for c in comms]
            # assert no duplicates
            self.assertEquals(3, len(ids))
            self.assertEquals(set(['comm-1', 'comm-2', 'comm-3']), set(ids))
            self.assertEquals(3, len(reader))
            self.assertTrue(hasattr(comms[0], 'sentenceForUUID'))
            self.assertTrue(hasattr(comms[1], 'sentenceForUUID'))
            self.assertTrue(hasattr(comms[2], 'sentenceForUUID'))
            batch_ids = [c.id for c in reader.batch(2)]
            # do this weird thing because set(['foo']) != set([u'foo'])
            self.assertTrue(
                ('comm-1' in batch_ids and 'comm-2' in batch_ids) or
                ('comm-1' in batch_ids and 'comm-3' in batch_ids) or
                ('comm-2' in batch_ids and 'comm-3' in batch_ids)
            )
            # assert data still there
            ids = [c.id for c in reader]
            self.assertEquals(set(['comm-1', 'comm-2', 'comm-3']), set(ids))
            self.assertEquals(3, redis_db.scard(key))

    def test_list(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.lpush(key, write_communication_to_buffer(self.comm1))
            redis_db.lpush(key, write_communication_to_buffer(self.comm2))
            redis_db.lpush(key, write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key, key_type='list')
            comms = [c for c in reader]
            ids = [c.id for c in comms]
            self.assertEquals(['comm-1', 'comm-2', 'comm-3'], ids)
            self.assertEquals(3, len(reader))
            self.assertEquals('comm-2', reader[1].id)
            self.assertTrue(hasattr(comms[0], 'sentenceForUUID'))
            self.assertTrue(hasattr(comms[1], 'sentenceForUUID'))
            self.assertTrue(hasattr(comms[2], 'sentenceForUUID'))
            # assert data still there
            ids = [c.id for c in reader]
            self.assertEquals(['comm-1', 'comm-2', 'comm-3'], ids)
            self.assertEquals(3, redis_db.llen(key))

    def test_hash(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.hset(key, self.comm1.uuid.uuidString,
                          write_communication_to_buffer(self.comm1))
            redis_db.hset(key, self.comm2.uuid.uuidString,
                          write_communication_to_buffer(self.comm2))
            redis_db.hset(key, self.comm3.uuid.uuidString,
                          write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key, key_type='hash')
            comms = [c for c in reader]
            ids = [c.id for c in comms]
            # assert no duplicates
            self.assertEquals(3, len(ids))
            self.assertEquals(set(['comm-1', 'comm-2', 'comm-3']), set(ids))
            self.assertEquals(3, len(reader))
            self.assertEquals('comm-2', reader[self.comm2.uuid.uuidString].id)
            self.assertTrue(hasattr(comms[0], 'sentenceForUUID'))
            self.assertTrue(hasattr(comms[1], 'sentenceForUUID'))
            self.assertTrue(hasattr(comms[2], 'sentenceForUUID'))
            # assert data still there
            ids = [c.id for c in reader]
            self.assertEquals(set(['comm-1', 'comm-2', 'comm-3']), set(ids))
            self.assertEquals(3, redis_db.hlen(key))

    def test_list_left_to_right(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.lpush(key, write_communication_to_buffer(self.comm1))
            redis_db.lpush(key, write_communication_to_buffer(self.comm2))
            redis_db.lpush(key, write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key, key_type='list',
                                              right_to_left=False)
            comms = [c for c in reader]
            ids = [c.id for c in comms]
            self.assertEquals(['comm-3', 'comm-2', 'comm-1'], ids)
            self.assertEquals(3, len(reader))
            self.assertEquals('comm-1', reader[2].id)
            self.assertTrue(hasattr(comms[0], 'sentenceForUUID'))
            self.assertTrue(hasattr(comms[1], 'sentenceForUUID'))
            self.assertTrue(hasattr(comms[2], 'sentenceForUUID'))
            # assert data still there
            ids = [c.id for c in reader]
            self.assertEquals(['comm-3', 'comm-2', 'comm-1'], ids)
            self.assertEquals(3, redis_db.llen(key))

    def test_set_implicit(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.sadd(key, write_communication_to_buffer(self.comm1))
            redis_db.sadd(key, write_communication_to_buffer(self.comm2))
            redis_db.sadd(key, write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key)
            ids = [c.id for c in reader]
            # assert no duplicates
            self.assertEquals(3, len(ids))
            self.assertEquals(set(['comm-1', 'comm-2', 'comm-3']), set(ids))

    def test_list_implicit(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.lpush(key, write_communication_to_buffer(self.comm1))
            redis_db.lpush(key, write_communication_to_buffer(self.comm2))
            redis_db.lpush(key, write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key)
            ids = [c.id for c in reader]
            self.assertEquals(['comm-1', 'comm-2', 'comm-3'], ids)

    def test_hash_implicit(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.hset(key, self.comm1.uuid.uuidString,
                          write_communication_to_buffer(self.comm1))
            redis_db.hset(key, self.comm2.uuid.uuidString,
                          write_communication_to_buffer(self.comm2))
            redis_db.hset(key, self.comm3.uuid.uuidString,
                          write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key)
            ids = [c.id for c in reader]
            # assert no duplicates
            self.assertEquals(3, len(ids))
            self.assertEquals(set(['comm-1', 'comm-2', 'comm-3']), set(ids))

    def test_set_empty(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            reader = RedisCommunicationReader(redis_db, key, key_type='set')
            self.assertEquals(0, len(reader))
            redis_db.sadd(key, write_communication_to_buffer(self.comm1))
            redis_db.sadd(key, write_communication_to_buffer(self.comm2))
            redis_db.sadd(key, write_communication_to_buffer(self.comm3))
            ids = [c.id for c in reader]
            # assert no duplicates
            self.assertEquals(3, len(ids))
            self.assertEquals(set(['comm-1', 'comm-2', 'comm-3']), set(ids))
            self.assertEquals(3, len(reader))

    def test_list_empty(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            reader = RedisCommunicationReader(redis_db, key, key_type='list')
            self.assertEquals(0, len(reader))
            redis_db.lpush(key, write_communication_to_buffer(self.comm1))
            redis_db.lpush(key, write_communication_to_buffer(self.comm2))
            redis_db.lpush(key, write_communication_to_buffer(self.comm3))
            ids = [c.id for c in reader]
            self.assertEquals(['comm-1', 'comm-2', 'comm-3'], ids)
            self.assertEquals(3, len(reader))

    def test_hash_empty(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            reader = RedisCommunicationReader(redis_db, key, key_type='hash')
            self.assertEquals(0, len(reader))
            redis_db.hset(key, self.comm1.uuid.uuidString,
                          write_communication_to_buffer(self.comm1))
            redis_db.hset(key, self.comm2.uuid.uuidString,
                          write_communication_to_buffer(self.comm2))
            redis_db.hset(key, self.comm3.uuid.uuidString,
                          write_communication_to_buffer(self.comm3))
            ids = [c.id for c in reader]
            # assert no duplicates
            self.assertEquals(3, len(ids))
            self.assertEquals(set(['comm-1', 'comm-2', 'comm-3']), set(ids))
            self.assertEquals(3, len(reader))

    def test_implicit_empty(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            with self.assertRaises(Exception):
                RedisCommunicationReader(redis_db, key)

    def test_set_no_add_references(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.sadd(key, write_communication_to_buffer(self.comm1))
            redis_db.sadd(key, write_communication_to_buffer(self.comm2))
            redis_db.sadd(key, write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key, key_type='set',
                                              add_references=False)
            comms = [c for c in reader]
            ids = [c.id for c in comms]
            # assert no duplicates
            self.assertEquals(3, len(ids))
            self.assertEquals(set(['comm-1', 'comm-2', 'comm-3']), set(ids))
            self.assertFalse(hasattr(comms[0], 'sentenceForUUID'))
            self.assertFalse(hasattr(comms[1], 'sentenceForUUID'))
            self.assertFalse(hasattr(comms[2], 'sentenceForUUID'))

    def test_list_no_add_references(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.lpush(key, write_communication_to_buffer(self.comm1))
            redis_db.lpush(key, write_communication_to_buffer(self.comm2))
            redis_db.lpush(key, write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key, key_type='list',
                                              add_references=False)
            comms = [c for c in reader]
            ids = [c.id for c in comms]
            self.assertEquals(['comm-1', 'comm-2', 'comm-3'], ids)
            self.assertFalse(hasattr(comms[0], 'sentenceForUUID'))
            self.assertFalse(hasattr(comms[1], 'sentenceForUUID'))
            self.assertFalse(hasattr(comms[2], 'sentenceForUUID'))

    def test_hash_no_add_references(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.hset(key, self.comm1.uuid.uuidString,
                          write_communication_to_buffer(self.comm1))
            redis_db.hset(key, self.comm2.uuid.uuidString,
                          write_communication_to_buffer(self.comm2))
            redis_db.hset(key, self.comm3.uuid.uuidString,
                          write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key, key_type='hash',
                                              add_references=False)
            comms = [c for c in reader]
            ids = [c.id for c in comms]
            # assert no duplicates
            self.assertEquals(3, len(ids))
            self.assertEquals(set(['comm-1', 'comm-2', 'comm-3']), set(ids))
            self.assertFalse(hasattr(comms[0], 'sentenceForUUID'))
            self.assertFalse(hasattr(comms[1], 'sentenceForUUID'))
            self.assertFalse(hasattr(comms[2], 'sentenceForUUID'))

    def test_set_pop(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.sadd(key, write_communication_to_buffer(self.comm1))
            redis_db.sadd(key, write_communication_to_buffer(self.comm2))
            redis_db.sadd(key, write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key, key_type='set',
                                              pop=True)
            it = iter(reader)
            ids = []
            ids.append(it.next().id)
            ids.append(it.next().id)
            self.assertEquals(1, redis_db.scard(key))
            ids.append(it.next().id)
            # assert no duplicates
            self.assertEquals(3, len(ids))
            self.assertEquals(set(['comm-1', 'comm-2', 'comm-3']), set(ids))
            # assert data is gone
            self.assertEquals([], [c.id for c in reader])
            self.assertFalse(redis_db.exists(key))
            with self.assertRaises(StopIteration):
                it.next()

    def test_list_pop(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.lpush(key, write_communication_to_buffer(self.comm1))
            redis_db.lpush(key, write_communication_to_buffer(self.comm2))
            redis_db.lpush(key, write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key, key_type='list',
                                              pop=True)
            it = iter(reader)
            ids = []
            ids.append(it.next().id)
            ids.append(it.next().id)
            self.assertEquals(1, redis_db.llen(key))
            ids.append(it.next().id)
            self.assertEquals(['comm-1', 'comm-2', 'comm-3'], ids)
            # assert data is gone
            self.assertEquals([], [c.id for c in reader])
            self.assertFalse(redis_db.exists(key))
            with self.assertRaises(StopIteration):
                it.next()

    def test_list_pop_left_to_right(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.lpush(key, write_communication_to_buffer(self.comm1))
            redis_db.lpush(key, write_communication_to_buffer(self.comm2))
            redis_db.lpush(key, write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key, key_type='list',
                                              pop=True, right_to_left=False)
            it = iter(reader)
            ids = []
            ids.append(it.next().id)
            ids.append(it.next().id)
            self.assertEquals(1, redis_db.llen(key))
            ids.append(it.next().id)
            self.assertEquals(['comm-3', 'comm-2', 'comm-1'], ids)
            # assert data is gone
            self.assertEquals([], [c.id for c in reader])
            self.assertFalse(redis_db.exists(key))
            with self.assertRaises(StopIteration):
                it.next()

    def test_list_block_pop(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.lpush(key, write_communication_to_buffer(self.comm1))
            redis_db.lpush(key, write_communication_to_buffer(self.comm2))
            redis_db.lpush(key, write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key, key_type='list',
                                              pop=True, block=True)
            it = iter(reader)
            ids = []
            ids.append(it.next().id)
            ids.append(it.next().id)
            self.assertEquals(1, redis_db.llen(key))
            ids.append(it.next().id)
            self.assertEquals(['comm-1', 'comm-2', 'comm-3'], ids)
            proc = Process(target=_add_comm_to_list,
                           args=(3, server.port, 'comm-4', key))
            proc.start()
            print 'Waiting for new comm to be added (3 sec)...'
            self.assertEquals('comm-4', iter(reader).next().id)
            proc.join()

    def test_list_block_pop_left_to_right(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.lpush(key, write_communication_to_buffer(self.comm1))
            redis_db.lpush(key, write_communication_to_buffer(self.comm2))
            redis_db.lpush(key, write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key, key_type='list',
                                              pop=True, block=True,
                                              right_to_left=False)
            it = iter(reader)
            ids = []
            ids.append(it.next().id)
            ids.append(it.next().id)
            self.assertEquals(1, redis_db.llen(key))
            ids.append(it.next().id)
            self.assertEquals(['comm-3', 'comm-2', 'comm-1'], ids)
            proc = Process(target=_add_comm_to_list,
                           args=(3, server.port, 'comm-4', key))
            proc.start()
            print 'Waiting for new comm to be added (3 sec)...'
            self.assertEquals('comm-4', iter(reader).next().id)
            proc.join()

    def test_list_block_pop_timeout(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.lpush(key, write_communication_to_buffer(self.comm1))
            redis_db.lpush(key, write_communication_to_buffer(self.comm2))
            redis_db.lpush(key, write_communication_to_buffer(self.comm3))
            reader = RedisCommunicationReader(redis_db, key, key_type='list',
                                              pop=True, block=True,
                                              block_timeout=1)
            it = iter(reader)
            ids = []
            ids.append(it.next().id)
            ids.append(it.next().id)
            self.assertEquals(1, redis_db.llen(key))
            ids.append(it.next().id)
            self.assertEquals(['comm-1', 'comm-2', 'comm-3'], ids)
            with self.assertRaises(StopIteration):
                print 'Waiting for timeout (1 sec)...'
                it.next()


class TestRedisCommunicationWriter(unittest.TestCase):

    def setUp(self):
        self.comm1 = create_comm('comm-1')
        self.comm2 = create_comm('comm-2')
        self.comm3 = create_comm('comm-3')
        self.buf1 = write_communication_to_buffer(self.comm1)
        self.buf2 = write_communication_to_buffer(self.comm2)
        self.buf3 = write_communication_to_buffer(self.comm3)

    def test_set(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            w = RedisCommunicationWriter(redis_db, key, key_type='set')
            w.write(self.comm1)
            self.assertEquals(1, redis_db.scard(key))
            self.assertEquals(self.buf1, redis_db.srandmember(key))
            w.write(self.comm2)
            w.write(self.comm3)
            self.assertEquals(3, redis_db.scard(key))

    def test_list(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            w = RedisCommunicationWriter(redis_db, key, key_type='list')
            w.write(self.comm1)
            self.assertEquals(1, redis_db.llen(key))
            self.assertEquals(self.buf1, redis_db.lindex(key, 0))
            w.write(self.comm2)
            w.write(self.comm3)
            self.assertEquals(3, redis_db.llen(key))
            self.assertEquals(self.buf1, redis_db.lindex(key, -1))
            self.assertEquals(self.buf2, redis_db.lindex(key, -2))
            self.assertEquals(self.buf3, redis_db.lindex(key, -3))

    def test_hash(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            w = RedisCommunicationWriter(redis_db, key, key_type='hash')
            w.write(self.comm1)
            self.assertEquals(1, redis_db.hlen(key))
            self.assertEquals(self.buf1, redis_db.hget(key, self.comm1.id))
            w.write(self.comm2)
            w.write(self.comm3)
            self.assertEquals(3, redis_db.hlen(key))
            self.assertEquals(self.buf1, redis_db.hget(key, self.comm1.id))
            self.assertEquals(self.buf2, redis_db.hget(key, self.comm2.id))
            self.assertEquals(self.buf3, redis_db.hget(key, self.comm3.id))

    def test_hash_uuid_key(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            w = RedisCommunicationWriter(
                redis_db, key, key_type='hash', uuid_hash_key=True)
            w.write(self.comm1)
            self.assertEquals(1, redis_db.hlen(key))
            self.assertEquals(self.buf1, redis_db.hget(
                key, self.comm1.uuid.uuidString))
            w.write(self.comm2)
            w.write(self.comm3)
            self.assertEquals(3, redis_db.hlen(key))
            self.assertEquals(self.buf1, redis_db.hget(
                key, self.comm1.uuid.uuidString))
            self.assertEquals(self.buf2, redis_db.hget(
                key, self.comm2.uuid.uuidString))
            self.assertEquals(self.buf3, redis_db.hget(
                key, self.comm3.uuid.uuidString))

    def test_list_left_to_right(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            w = RedisCommunicationWriter(redis_db, key, key_type='list',
                                         right_to_left=False)
            w.write(self.comm1)
            self.assertEquals(1, redis_db.llen(key))
            self.assertEquals(self.buf1, redis_db.lindex(key, 0))
            w.write(self.comm2)
            w.write(self.comm3)
            self.assertEquals(3, redis_db.llen(key))
            self.assertEquals(self.buf3, redis_db.lindex(key, -1))
            self.assertEquals(self.buf2, redis_db.lindex(key, -2))
            self.assertEquals(self.buf1, redis_db.lindex(key, -3))

    def test_set_implicit(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.sadd(key, self.buf1)
            w = RedisCommunicationWriter(redis_db, key)
            w.write(self.comm2)
            w.write(self.comm3)
            self.assertEquals(3, redis_db.scard(key))

    def test_list_implicit(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.lpush(key, self.buf1)
            w = RedisCommunicationWriter(redis_db, key)
            w.write(self.comm2)
            w.write(self.comm3)
            self.assertEquals(3, redis_db.llen(key))
            self.assertEquals(self.buf1, redis_db.lindex(key, -1))
            self.assertEquals(self.buf2, redis_db.lindex(key, -2))
            self.assertEquals(self.buf3, redis_db.lindex(key, -3))

    def test_hash_implicit(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            redis_db.hset(key, self.comm1.id, self.buf1)
            w = RedisCommunicationWriter(redis_db, key)
            w.write(self.comm2)
            w.write(self.comm3)
            self.assertEquals(3, redis_db.hlen(key))
            self.assertEquals(self.buf1, redis_db.hget(key, self.comm1.id))
            self.assertEquals(self.buf2, redis_db.hget(key, self.comm2.id))
            self.assertEquals(self.buf3, redis_db.hget(key, self.comm3.id))

    def test_implicit_empty(self):
        key = 'dataset'
        with RedisServer(loglevel='warning') as server:
            redis_db = Redis(port=server.port)
            with self.assertRaises(Exception):
                RedisCommunicationWriter(redis_db, key)



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.9.3)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:new_style,utf8strings,coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException
import concrete.services.Service
import logging
from ttypes import *
from thrift.Thrift import TProcessor
from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
  from thrift.protocol import fastbinary
except:
  fastbinary = None


class Iface(concrete.services.Service.Iface):
  """
  The search proxy service provides a single interface to multiple search providers
  """
  def search(self, query, provider):
    """
    Specify the search provider when performing a search

    Parameters:
     - query
     - provider
    """
    pass

  def getProviders(self):
    """
    Get a list of search providers behind the proxy
    """
    pass

  def getCapabilities(self, provider):
    """
    Get a list of search type and language pairs for a search provider

    Parameters:
     - provider
    """
    pass

  def getCorpora(self, provider):
    """
    Get a corpus list for a search provider

    Parameters:
     - provider
    """
    pass


class Client(concrete.services.Service.Client, Iface):
  """
  The search proxy service provides a single interface to multiple search providers
  """
  def __init__(self, iprot, oprot=None):
    concrete.services.Service.Client.__init__(self, iprot, oprot)

  def search(self, query, provider):
    """
    Specify the search provider when performing a search

    Parameters:
     - query
     - provider
    """
    self.send_search(query, provider)
    return self.recv_search()

  def send_search(self, query, provider):
    self._oprot.writeMessageBegin('search', TMessageType.CALL, self._seqid)
    args = search_args()
    args.query = query
    args.provider = provider
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_search(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = search_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    if result.ex is not None:
      raise result.ex
    raise TApplicationException(TApplicationException.MISSING_RESULT, "search failed: unknown result")

  def getProviders(self):
    """
    Get a list of search providers behind the proxy
    """
    self.send_getProviders()
    return self.recv_getProviders()

  def send_getProviders(self):
    self._oprot.writeMessageBegin('getProviders', TMessageType.CALL, self._seqid)
    args = getProviders_args()
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_getProviders(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = getProviders_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    if result.ex is not None:
      raise result.ex
    raise TApplicationException(TApplicationException.MISSING_RESULT, "getProviders failed: unknown result")

  def getCapabilities(self, provider):
    """
    Get a list of search type and language pairs for a search provider

    Parameters:
     - provider
    """
    self.send_getCapabilities(provider)
    return self.recv_getCapabilities()

  def send_getCapabilities(self, provider):
    self._oprot.writeMessageBegin('getCapabilities', TMessageType.CALL, self._seqid)
    args = getCapabilities_args()
    args.provider = provider
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_getCapabilities(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = getCapabilities_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    if result.ex is not None:
      raise result.ex
    raise TApplicationException(TApplicationException.MISSING_RESULT, "getCapabilities failed: unknown result")

  def getCorpora(self, provider):
    """
    Get a corpus list for a search provider

    Parameters:
     - provider
    """
    self.send_getCorpora(provider)
    return self.recv_getCorpora()

  def send_getCorpora(self, provider):
    self._oprot.writeMessageBegin('getCorpora', TMessageType.CALL, self._seqid)
    args = getCorpora_args()
    args.provider = provider
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_getCorpora(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = getCorpora_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    if result.ex is not None:
      raise result.ex
    raise TApplicationException(TApplicationException.MISSING_RESULT, "getCorpora failed: unknown result")


class Processor(concrete.services.Service.Processor, Iface, TProcessor):
  def __init__(self, handler):
    concrete.services.Service.Processor.__init__(self, handler)
    self._processMap["search"] = Processor.process_search
    self._processMap["getProviders"] = Processor.process_getProviders
    self._processMap["getCapabilities"] = Processor.process_getCapabilities
    self._processMap["getCorpora"] = Processor.process_getCorpora

  def process(self, iprot, oprot):
    (name, type, seqid) = iprot.readMessageBegin()
    if name not in self._processMap:
      iprot.skip(TType.STRUCT)
      iprot.readMessageEnd()
      x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
      oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
      x.write(oprot)
      oprot.writeMessageEnd()
      oprot.trans.flush()
      return
    else:
      self._processMap[name](self, seqid, iprot, oprot)
    return True

  def process_search(self, seqid, iprot, oprot):
    args = search_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = search_result()
    try:
      result.success = self._handler.search(args.query, args.provider)
      msg_type = TMessageType.REPLY
    except (TTransport.TTransportException, KeyboardInterrupt, SystemExit):
      raise
    except concrete.services.ttypes.ServicesException as ex:
      msg_type = TMessageType.REPLY
      result.ex = ex
    except Exception as ex:
      msg_type = TMessageType.EXCEPTION
      logging.exception(ex)
      result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
    oprot.writeMessageBegin("search", msg_type, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_getProviders(self, seqid, iprot, oprot):
    args = getProviders_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = getProviders_result()
    try:
      result.success = self._handler.getProviders()
      msg_type = TMessageType.REPLY
    except (TTransport.TTransportException, KeyboardInterrupt, SystemExit):
      raise
    except concrete.services.ttypes.ServicesException as ex:
      msg_type = TMessageType.REPLY
      result.ex = ex
    except Exception as ex:
      msg_type = TMessageType.EXCEPTION
      logging.exception(ex)
      result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
    oprot.writeMessageBegin("getProviders", msg_type, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_getCapabilities(self, seqid, iprot, oprot):
    args = getCapabilities_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = getCapabilities_result()
    try:
      result.success = self._handler.getCapabilities(args.provider)
      msg_type = TMessageType.REPLY
    except (TTransport.TTransportException, KeyboardInterrupt, SystemExit):
      raise
    except concrete.services.ttypes.ServicesException as ex:
      msg_type = TMessageType.REPLY
      result.ex = ex
    except Exception as ex:
      msg_type = TMessageType.EXCEPTION
      logging.exception(ex)
      result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
    oprot.writeMessageBegin("getCapabilities", msg_type, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_getCorpora(self, seqid, iprot, oprot):
    args = getCorpora_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = getCorpora_result()
    try:
      result.success = self._handler.getCorpora(args.provider)
      msg_type = TMessageType.REPLY
    except (TTransport.TTransportException, KeyboardInterrupt, SystemExit):
      raise
    except concrete.services.ttypes.ServicesException as ex:
      msg_type = TMessageType.REPLY
      result.ex = ex
    except Exception as ex:
      msg_type = TMessageType.EXCEPTION
      logging.exception(ex)
      result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
    oprot.writeMessageBegin("getCorpora", msg_type, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()


# HELPER FUNCTIONS AND STRUCTURES

class search_args(object):
  """
  Attributes:
   - query
   - provider
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'query', (SearchQuery, SearchQuery.thrift_spec), None, ), # 1
    (2, TType.STRING, 'provider', None, None, ), # 2
  )

  def __init__(self, query=None, provider=None,):
    self.query = query
    self.provider = provider

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.query = SearchQuery()
          self.query.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.provider = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('search_args')
    if self.query is not None:
      oprot.writeFieldBegin('query', TType.STRUCT, 1)
      self.query.write(oprot)
      oprot.writeFieldEnd()
    if self.provider is not None:
      oprot.writeFieldBegin('provider', TType.STRING, 2)
      oprot.writeString(self.provider.encode('utf-8'))
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.query)
    value = (value * 31) ^ hash(self.provider)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class search_result(object):
  """
  Attributes:
   - success
   - ex
  """

  thrift_spec = (
    (0, TType.STRUCT, 'success', (SearchResults, SearchResults.thrift_spec), None, ), # 0
    (1, TType.STRUCT, 'ex', (concrete.services.ttypes.ServicesException, concrete.services.ttypes.ServicesException.thrift_spec), None, ), # 1
  )

  def __init__(self, success=None, ex=None,):
    self.success = success
    self.ex = ex

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.STRUCT:
          self.success = SearchResults()
          self.success.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 1:
        if ftype == TType.STRUCT:
          self.ex = concrete.services.ttypes.ServicesException()
          self.ex.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('search_result')
    if self.success is not None:
      oprot.writeFieldBegin('success', TType.STRUCT, 0)
      self.success.write(oprot)
      oprot.writeFieldEnd()
    if self.ex is not None:
      oprot.writeFieldBegin('ex', TType.STRUCT, 1)
      self.ex.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    value = (value * 31) ^ hash(self.ex)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class getProviders_args(object):

  thrift_spec = (
  )

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('getProviders_args')
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class getProviders_result(object):
  """
  Attributes:
   - success
   - ex
  """

  thrift_spec = (
    (0, TType.LIST, 'success', (TType.STRING,None), None, ), # 0
    (1, TType.STRUCT, 'ex', (concrete.services.ttypes.ServicesException, concrete.services.ttypes.ServicesException.thrift_spec), None, ), # 1
  )

  def __init__(self, success=None, ex=None,):
    self.success = success
    self.ex = ex

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.LIST:
          self.success = []
          (_etype45, _size42) = iprot.readListBegin()
          for _i46 in xrange(_size42):
            _elem47 = iprot.readString().decode('utf-8')
            self.success.append(_elem47)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 1:
        if ftype == TType.STRUCT:
          self.ex = concrete.services.ttypes.ServicesException()
          self.ex.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('getProviders_result')
    if self.success is not None:
      oprot.writeFieldBegin('success', TType.LIST, 0)
      oprot.writeListBegin(TType.STRING, len(self.success))
      for iter48 in self.success:
        oprot.writeString(iter48.encode('utf-8'))
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.ex is not None:
      oprot.writeFieldBegin('ex', TType.STRUCT, 1)
      self.ex.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    value = (value * 31) ^ hash(self.ex)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class getCapabilities_args(object):
  """
  Attributes:
   - provider
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'provider', None, None, ), # 1
  )

  def __init__(self, provider=None,):
    self.provider = provider

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.provider = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('getCapabilities_args')
    if self.provider is not None:
      oprot.writeFieldBegin('provider', TType.STRING, 1)
      oprot.writeString(self.provider.encode('utf-8'))
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.provider)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class getCapabilities_result(object):
  """
  Attributes:
   - success
   - ex
  """

  thrift_spec = (
    (0, TType.LIST, 'success', (TType.STRUCT,(SearchCapability, SearchCapability.thrift_spec)), None, ), # 0
    (1, TType.STRUCT, 'ex', (concrete.services.ttypes.ServicesException, concrete.services.ttypes.ServicesException.thrift_spec), None, ), # 1
  )

  def __init__(self, success=None, ex=None,):
    self.success = success
    self.ex = ex

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.LIST:
          self.success = []
          (_etype52, _size49) = iprot.readListBegin()
          for _i53 in xrange(_size49):
            _elem54 = SearchCapability()
            _elem54.read(iprot)
            self.success.append(_elem54)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 1:
        if ftype == TType.STRUCT:
          self.ex = concrete.services.ttypes.ServicesException()
          self.ex.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('getCapabilities_result')
    if self.success is not None:
      oprot.writeFieldBegin('success', TType.LIST, 0)
      oprot.writeListBegin(TType.STRUCT, len(self.success))
      for iter55 in self.success:
        iter55.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.ex is not None:
      oprot.writeFieldBegin('ex', TType.STRUCT, 1)
      self.ex.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    value = (value * 31) ^ hash(self.ex)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class getCorpora_args(object):
  """
  Attributes:
   - provider
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'provider', None, None, ), # 1
  )

  def __init__(self, provider=None,):
    self.provider = provider

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.provider = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('getCorpora_args')
    if self.provider is not None:
      oprot.writeFieldBegin('provider', TType.STRING, 1)
      oprot.writeString(self.provider.encode('utf-8'))
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.provider)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class getCorpora_result(object):
  """
  Attributes:
   - success
   - ex
  """

  thrift_spec = (
    (0, TType.LIST, 'success', (TType.STRING,None), None, ), # 0
    (1, TType.STRUCT, 'ex', (concrete.services.ttypes.ServicesException, concrete.services.ttypes.ServicesException.thrift_spec), None, ), # 1
  )

  def __init__(self, success=None, ex=None,):
    self.success = success
    self.ex = ex

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.LIST:
          self.success = []
          (_etype59, _size56) = iprot.readListBegin()
          for _i60 in xrange(_size56):
            _elem61 = iprot.readString().decode('utf-8')
            self.success.append(_elem61)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 1:
        if ftype == TType.STRUCT:
          self.ex = concrete.services.ttypes.ServicesException()
          self.ex.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('getCorpora_result')
    if self.success is not None:
      oprot.writeFieldBegin('success', TType.LIST, 0)
      oprot.writeListBegin(TType.STRING, len(self.success))
      for iter62 in self.success:
        oprot.writeString(iter62.encode('utf-8'))
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.ex is not None:
      oprot.writeFieldBegin('ex', TType.STRUCT, 1)
      self.ex.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    value = (value * 31) ^ hash(self.ex)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.9.3)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:new_style,utf8strings,coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException
import concrete.metadata.ttypes
import concrete.language.ttypes
import concrete.structure.ttypes
import concrete.entities.ttypes
import concrete.situations.ttypes
import concrete.exceptions.ttypes
import concrete.email.ttypes
import concrete.twitter.ttypes
import concrete.audio.ttypes
import concrete.communication.ttypes


from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
  from thrift.protocol import fastbinary
except:
  fastbinary = None





# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.9.3)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:new_style,utf8strings,coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException
import concrete.services.ttypes
import concrete.structure.ttypes
import concrete.uuid.ttypes
import concrete.metadata.ttypes


from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
  from thrift.protocol import fastbinary
except:
  fastbinary = None


class SearchType(object):
  """
  What are we searching over
  """
  COMMUNICATIONS = 0
  SECTIONS = 1
  SENTENCES = 2
  ENTITIES = 3
  ENTITY_MENTIONS = 4
  SITUATIONS = 5
  SITUATION_MENTIONS = 6

  _VALUES_TO_NAMES = {
    0: "COMMUNICATIONS",
    1: "SECTIONS",
    2: "SENTENCES",
    3: "ENTITIES",
    4: "ENTITY_MENTIONS",
    5: "SITUATIONS",
    6: "SITUATION_MENTIONS",
  }

  _NAMES_TO_VALUES = {
    "COMMUNICATIONS": 0,
    "SECTIONS": 1,
    "SENTENCES": 2,
    "ENTITIES": 3,
    "ENTITY_MENTIONS": 4,
    "SITUATIONS": 5,
    "SITUATION_MENTIONS": 6,
  }

class SearchFeedback(object):
  """
  Feedback values
  """
  NEGATIVE = -1
  NONE = 0
  POSITIVE = 1

  _VALUES_TO_NAMES = {
    -1: "NEGATIVE",
    0: "NONE",
    1: "POSITIVE",
  }

  _NAMES_TO_VALUES = {
    "NEGATIVE": -1,
    "NONE": 0,
    "POSITIVE": 1,
  }


class SearchCapability(object):
  """
  A search provider describes its capabilities with a list of search type and language pairs.

  Attributes:
   - type: A type of search supported by the search provider
   - lang: Language that the search provider supports.
  Use ISO 639-2/T three letter codes.
  """

  thrift_spec = (
    None, # 0
    (1, TType.I32, 'type', None, None, ), # 1
    (2, TType.STRING, 'lang', None, None, ), # 2
  )

  def __init__(self, type=None, lang=None,):
    self.type = type
    self.lang = lang

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.I32:
          self.type = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.lang = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('SearchCapability')
    if self.type is not None:
      oprot.writeFieldBegin('type', TType.I32, 1)
      oprot.writeI32(self.type)
      oprot.writeFieldEnd()
    if self.lang is not None:
      oprot.writeFieldBegin('lang', TType.STRING, 2)
      oprot.writeString(self.lang.encode('utf-8'))
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.type is None:
      raise TProtocol.TProtocolException(message='Required field type is unset!')
    if self.lang is None:
      raise TProtocol.TProtocolException(message='Required field lang is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.type)
    value = (value * 31) ^ hash(self.lang)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class SearchQuery(object):
  """
  Wrapper for information relevant to a (possibly structured) search.

  Attributes:
   - terms: Individual words, or multiword phrases, e.g., 'dog', 'blue
  cheese'.  It is the responsibility of the implementation of
  Search* to tokenize multiword phrases, if so-desired.  Further,
  an implementation may choose to support advanced features such as
  wildcards, e.g.: 'blue*'.  This specification makes no
  committment as to the internal structure of keywords and their
  semantics: that is the responsibility of the individual
  implementation.
   - questions: e.g., "what is the capital of spain?"

  questions is a list in order that possibly different phrasings of
  the question can be included, e.g.: "what is the name of spain's
  capital?"
   - communicationId: Refers to an optional communication that can provide context for the query.
   - tokens: Refers to a sequence of tokens in the communication referenced by communicationId.
   - rawQuery: The input from the user provided in the search box, unmodified
   - auths: optional authorization mechanism
   - userId: Identifies the user who submitted the search query
   - name: Human readable name of the query.
   - labels: Properties of the query or user.
  These labels can be used to group queries and results by a domain or group of
  users for training. An example usage would be assigning the geographical region
  as a label ("spain"). User labels could be based on organizational units ("hltcoe").
   - type: This search is over this type of data (communications, sentences, entities)
   - lang: The language of the corpus that the user wants to search.
  Use ISO 639-2/T three letter codes.
   - corpus: An identifier of the corpus that the search is to be performed over.
  """

  thrift_spec = (
    None, # 0
    (1, TType.LIST, 'terms', (TType.STRING,None), None, ), # 1
    (2, TType.LIST, 'questions', (TType.STRING,None), None, ), # 2
    (3, TType.STRING, 'communicationId', None, None, ), # 3
    (4, TType.STRUCT, 'tokens', (concrete.structure.ttypes.TokenRefSequence, concrete.structure.ttypes.TokenRefSequence.thrift_spec), None, ), # 4
    (5, TType.STRING, 'rawQuery', None, None, ), # 5
    (6, TType.STRING, 'auths', None, None, ), # 6
    (7, TType.STRING, 'userId', None, None, ), # 7
    (8, TType.STRING, 'name', None, None, ), # 8
    (9, TType.LIST, 'labels', (TType.STRING,None), None, ), # 9
    (10, TType.I32, 'type', None, None, ), # 10
    (11, TType.STRING, 'lang', None, None, ), # 11
    (12, TType.STRING, 'corpus', None, None, ), # 12
  )

  def __init__(self, terms=None, questions=None, communicationId=None, tokens=None, rawQuery=None, auths=None, userId=None, name=None, labels=None, type=None, lang=None, corpus=None,):
    self.terms = terms
    self.questions = questions
    self.communicationId = communicationId
    self.tokens = tokens
    self.rawQuery = rawQuery
    self.auths = auths
    self.userId = userId
    self.name = name
    self.labels = labels
    self.type = type
    self.lang = lang
    self.corpus = corpus

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.LIST:
          self.terms = []
          (_etype3, _size0) = iprot.readListBegin()
          for _i4 in xrange(_size0):
            _elem5 = iprot.readString().decode('utf-8')
            self.terms.append(_elem5)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.LIST:
          self.questions = []
          (_etype9, _size6) = iprot.readListBegin()
          for _i10 in xrange(_size6):
            _elem11 = iprot.readString().decode('utf-8')
            self.questions.append(_elem11)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.STRING:
          self.communicationId = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.STRUCT:
          self.tokens = concrete.structure.ttypes.TokenRefSequence()
          self.tokens.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 5:
        if ftype == TType.STRING:
          self.rawQuery = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      elif fid == 6:
        if ftype == TType.STRING:
          self.auths = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      elif fid == 7:
        if ftype == TType.STRING:
          self.userId = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      elif fid == 8:
        if ftype == TType.STRING:
          self.name = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      elif fid == 9:
        if ftype == TType.LIST:
          self.labels = []
          (_etype15, _size12) = iprot.readListBegin()
          for _i16 in xrange(_size12):
            _elem17 = iprot.readString().decode('utf-8')
            self.labels.append(_elem17)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 10:
        if ftype == TType.I32:
          self.type = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 11:
        if ftype == TType.STRING:
          self.lang = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      elif fid == 12:
        if ftype == TType.STRING:
          self.corpus = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('SearchQuery')
    if self.terms is not None:
      oprot.writeFieldBegin('terms', TType.LIST, 1)
      oprot.writeListBegin(TType.STRING, len(self.terms))
      for iter18 in self.terms:
        oprot.writeString(iter18.encode('utf-8'))
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.questions is not None:
      oprot.writeFieldBegin('questions', TType.LIST, 2)
      oprot.writeListBegin(TType.STRING, len(self.questions))
      for iter19 in self.questions:
        oprot.writeString(iter19.encode('utf-8'))
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.communicationId is not None:
      oprot.writeFieldBegin('communicationId', TType.STRING, 3)
      oprot.writeString(self.communicationId.encode('utf-8'))
      oprot.writeFieldEnd()
    if self.tokens is not None:
      oprot.writeFieldBegin('tokens', TType.STRUCT, 4)
      self.tokens.write(oprot)
      oprot.writeFieldEnd()
    if self.rawQuery is not None:
      oprot.writeFieldBegin('rawQuery', TType.STRING, 5)
      oprot.writeString(self.rawQuery.encode('utf-8'))
      oprot.writeFieldEnd()
    if self.auths is not None:
      oprot.writeFieldBegin('auths', TType.STRING, 6)
      oprot.writeString(self.auths.encode('utf-8'))
      oprot.writeFieldEnd()
    if self.userId is not None:
      oprot.writeFieldBegin('userId', TType.STRING, 7)
      oprot.writeString(self.userId.encode('utf-8'))
      oprot.writeFieldEnd()
    if self.name is not None:
      oprot.writeFieldBegin('name', TType.STRING, 8)
      oprot.writeString(self.name.encode('utf-8'))
      oprot.writeFieldEnd()
    if self.labels is not None:
      oprot.writeFieldBegin('labels', TType.LIST, 9)
      oprot.writeListBegin(TType.STRING, len(self.labels))
      for iter20 in self.labels:
        oprot.writeString(iter20.encode('utf-8'))
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.type is not None:
      oprot.writeFieldBegin('type', TType.I32, 10)
      oprot.writeI32(self.type)
      oprot.writeFieldEnd()
    if self.lang is not None:
      oprot.writeFieldBegin('lang', TType.STRING, 11)
      oprot.writeString(self.lang.encode('utf-8'))
      oprot.writeFieldEnd()
    if self.corpus is not None:
      oprot.writeFieldBegin('corpus', TType.STRING, 12)
      oprot.writeString(self.corpus.encode('utf-8'))
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.type is None:
      raise TProtocol.TProtocolException(message='Required field type is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.terms)
    value = (value * 31) ^ hash(self.questions)
    value = (value * 31) ^ hash(self.communicationId)
    value = (value * 31) ^ hash(self.tokens)
    value = (value * 31) ^ hash(self.rawQuery)
    value = (value * 31) ^ hash(self.auths)
    value = (value * 31) ^ hash(self.userId)
    value = (value * 31) ^ hash(self.name)
    value = (value * 31) ^ hash(self.labels)
    value = (value * 31) ^ hash(self.type)
    value = (value * 31) ^ hash(self.lang)
    value = (value * 31) ^ hash(self.corpus)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class SearchResult(object):
  """
  An individual element returned from a search.  Most/all methods
  will return a communicationId, possibly with an associated score.
  For example if the target element type of the search is Sentence
  then the sentenceId field should be populated.

  Attributes:
   - communicationId
   - sentenceId: The UUID of the returned sentence, which appears in the
  communication referenced by communicationId.
   - score: Values are not restricted in range (e.g., do not have to be
  within [0,1]).  Higher is better.

   - tokens: If the Search is meant to result in a tokenRefSequence, this is
  that result.  Otherwise, this field may be optionally populated
  in order to provide a hint to the client as to where to center a
  visualization, or the extraction of context, etc.
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRING, 'communicationId', None, None, ), # 1
    (2, TType.STRUCT, 'sentenceId', (concrete.uuid.ttypes.UUID, concrete.uuid.ttypes.UUID.thrift_spec), None, ), # 2
    (3, TType.DOUBLE, 'score', None, None, ), # 3
    (4, TType.STRUCT, 'tokens', (concrete.structure.ttypes.TokenRefSequence, concrete.structure.ttypes.TokenRefSequence.thrift_spec), None, ), # 4
  )

  def __init__(self, communicationId=None, sentenceId=None, score=None, tokens=None,):
    self.communicationId = communicationId
    self.sentenceId = sentenceId
    self.score = score
    self.tokens = tokens

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRING:
          self.communicationId = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRUCT:
          self.sentenceId = concrete.uuid.ttypes.UUID()
          self.sentenceId.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.DOUBLE:
          self.score = iprot.readDouble()
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.STRUCT:
          self.tokens = concrete.structure.ttypes.TokenRefSequence()
          self.tokens.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('SearchResult')
    if self.communicationId is not None:
      oprot.writeFieldBegin('communicationId', TType.STRING, 1)
      oprot.writeString(self.communicationId.encode('utf-8'))
      oprot.writeFieldEnd()
    if self.sentenceId is not None:
      oprot.writeFieldBegin('sentenceId', TType.STRUCT, 2)
      self.sentenceId.write(oprot)
      oprot.writeFieldEnd()
    if self.score is not None:
      oprot.writeFieldBegin('score', TType.DOUBLE, 3)
      oprot.writeDouble(self.score)
      oprot.writeFieldEnd()
    if self.tokens is not None:
      oprot.writeFieldBegin('tokens', TType.STRUCT, 4)
      self.tokens.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.communicationId)
    value = (value * 31) ^ hash(self.sentenceId)
    value = (value * 31) ^ hash(self.score)
    value = (value * 31) ^ hash(self.tokens)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class SearchResults(object):
  """
  Single wrapper for results from all the various Search* services.

  Attributes:
   - uuid: Unique identifier for the results of this search.
   - searchQuery: The query that led to this result.
  Useful for capturing feedback or building training data.
   - searchResults: The list is assumed sorted best to worst, which should be
  reflected by the values contained in the score field of each
  SearchResult, if that field is populated.
   - metadata: The system that provided the response: likely use case for
  populating this field is for building training data.  Presumably
  a system will not need/want to return this object in live use.
   - lang: The dominant language of the search results.
  Use ISO 639-2/T three letter codes.
  Search providers should set this when possible to support downstream processing.
  Do not set if it is not known.
  If multilingual, use the string "multilingual".
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'uuid', (concrete.uuid.ttypes.UUID, concrete.uuid.ttypes.UUID.thrift_spec), None, ), # 1
    (2, TType.STRUCT, 'searchQuery', (SearchQuery, SearchQuery.thrift_spec), None, ), # 2
    (3, TType.LIST, 'searchResults', (TType.STRUCT,(SearchResult, SearchResult.thrift_spec)), None, ), # 3
    (4, TType.STRUCT, 'metadata', (concrete.metadata.ttypes.AnnotationMetadata, concrete.metadata.ttypes.AnnotationMetadata.thrift_spec), None, ), # 4
    (5, TType.STRING, 'lang', None, None, ), # 5
  )

  def __init__(self, uuid=None, searchQuery=None, searchResults=None, metadata=None, lang=None,):
    self.uuid = uuid
    self.searchQuery = searchQuery
    self.searchResults = searchResults
    self.metadata = metadata
    self.lang = lang

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.uuid = concrete.uuid.ttypes.UUID()
          self.uuid.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRUCT:
          self.searchQuery = SearchQuery()
          self.searchQuery.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.LIST:
          self.searchResults = []
          (_etype24, _size21) = iprot.readListBegin()
          for _i25 in xrange(_size21):
            _elem26 = SearchResult()
            _elem26.read(iprot)
            self.searchResults.append(_elem26)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.STRUCT:
          self.metadata = concrete.metadata.ttypes.AnnotationMetadata()
          self.metadata.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 5:
        if ftype == TType.STRING:
          self.lang = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('SearchResults')
    if self.uuid is not None:
      oprot.writeFieldBegin('uuid', TType.STRUCT, 1)
      self.uuid.write(oprot)
      oprot.writeFieldEnd()
    if self.searchQuery is not None:
      oprot.writeFieldBegin('searchQuery', TType.STRUCT, 2)
      self.searchQuery.write(oprot)
      oprot.writeFieldEnd()
    if self.searchResults is not None:
      oprot.writeFieldBegin('searchResults', TType.LIST, 3)
      oprot.writeListBegin(TType.STRUCT, len(self.searchResults))
      for iter27 in self.searchResults:
        iter27.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    if self.metadata is not None:
      oprot.writeFieldBegin('metadata', TType.STRUCT, 4)
      self.metadata.write(oprot)
      oprot.writeFieldEnd()
    if self.lang is not None:
      oprot.writeFieldBegin('lang', TType.STRING, 5)
      oprot.writeString(self.lang.encode('utf-8'))
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.uuid is None:
      raise TProtocol.TProtocolException(message='Required field uuid is unset!')
    if self.searchQuery is None:
      raise TProtocol.TProtocolException(message='Required field searchQuery is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.uuid)
    value = (value * 31) ^ hash(self.searchQuery)
    value = (value * 31) ^ hash(self.searchResults)
    value = (value * 31) ^ hash(self.metadata)
    value = (value * 31) ^ hash(self.lang)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)



import socket


def find_port():
    '''
    Find and return an available TCP port.

    >>> find_port() > 1023
    True
    '''
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.bind(('', 0))
    addr = s.getsockname()
    port = addr[1]
    s.close()
    return port



# -*- coding: utf-8 -*-


from pytest import fixture
from concrete.util.file_io import CommunicationReader
from concrete.validate import validate_communication
import os
from subprocess import Popen, PIPE
from tempfile import mkstemp


def assert_first_comm(comm):
    assert comm.id == '238426131689242624'
    assert comm.startTime == 1345680194
    assert comm.endTime == 1345680194
    assert validate_communication(comm)


def assert_second_comm(comm):
    assert comm.id == '238426131689242625'
    assert comm.startTime == 1345680195
    assert comm.endTime == 1345680195
    assert validate_communication(comm)


@fixture
def log_conf(request):
    (fd, log_path) = mkstemp()
    os.close(fd)

    (fd, conf_path) = mkstemp()
    os.close(fd)
    with open(conf_path, 'w') as f:
        f.write('''{
  "version": 1,
  "root": {
    "level": "INFO",
    "handlers": ["file"]
  },
  "formatters": {
    "medium": {
      "format": "%%(asctime)-15s %%(levelname)s: %%(message)s"
    }
  },
  "handlers": {
    "file": {
      "class": "logging.FileHandler",
      "formatter": "medium",
      "filename": "%s"
    }
  }
}
''' % log_path)

    def _remove():
        if os.path.exists(conf_path):
            os.remove(conf_path)
        if os.path.exists(log_path):
            os.remove(log_path)

    request.addfinalizer(_remove)
    return (conf_path, log_path)


@fixture
def output_file(request):
    (fd, path) = mkstemp()
    os.close(fd)

    def _remove():
        if os.path.exists(path):
            os.remove(path)

    request.addfinalizer(_remove)
    return path


def test_tweets2concrete(output_file):
    p = Popen([
        'scripts/tweets2concrete.py',
        'tests/testdata/tweets.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = it.next()
    assert_first_comm(comm)

    (comm, _) = it.next()
    assert_second_comm(comm)

    try:
        it.next()
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_stdin(output_file):
    p = Popen([
        'scripts/tweets2concrete.py',
        '-',
        output_file
    ], stdin=PIPE, stdout=PIPE, stderr=PIPE)
    with open('tests/testdata/tweets.json', 'rb') as f:
        (stdout, stderr) = p.communicate(f.read())
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = it.next()
    assert_first_comm(comm)

    (comm, _) = it.next()
    assert_second_comm(comm)

    try:
        it.next()
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_stdout(output_file):
    p = Popen([
        'scripts/tweets2concrete.py',
        'tests/testdata/tweets.json',
        '-'
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    with open(output_file, 'wb') as f:
        f.write(stdout)

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = it.next()
    assert_first_comm(comm)

    (comm, _) = it.next()
    assert_second_comm(comm)

    try:
        it.next()
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_multiproc(output_file):
    p = Popen([
        'scripts/tweets2concrete.py',
        '--num-proc', '2',
        'tests/testdata/tweets.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = it.next()
    assert_first_comm(comm)

    (comm, _) = it.next()
    assert_second_comm(comm)

    try:
        it.next()
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_log_every(output_file):
    p = Popen([
        'scripts/tweets2concrete.py',
        '--log-level', 'INFO',
        '--log-interval', '1',
        'tests/testdata/tweets.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0
    assert len([line
                for line in stderr.strip().split('\n')
                if 'INFO' in line]) >= 2

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = it.next()
    assert_first_comm(comm)

    (comm, _) = it.next()
    assert_second_comm(comm)

    try:
        it.next()
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_unicode(output_file):
    p = Popen([
        'scripts/tweets2concrete.py',
        'tests/testdata/tweets.unicode.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = it.next()
    assert_first_comm(comm)
    assert validate_communication(comm)

    (comm, _) = it.next()
    assert_second_comm(comm)

    try:
        it.next()
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_gz(output_file):
    p = Popen([
        'scripts/tweets2concrete.py',
        'tests/testdata/tweets.json.gz',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = it.next()
    assert_first_comm(comm)

    (comm, _) = it.next()
    assert_second_comm(comm)

    try:
        it.next()
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_incomplete_gz(output_file):
    p = Popen([
        'scripts/tweets2concrete.py',
        '--catch-ioerror',
        'tests/testdata/tweets.json.incomplete.gz',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = it.next()
    assert_first_comm(comm)

    (comm, _) = it.next()
    assert_second_comm(comm)

    try:
        it.next()
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_incomplete_gz_multiproc(output_file):
    p = Popen([
        'scripts/tweets2concrete.py',
        '--num-proc', '2',
        '--catch-ioerror',
        'tests/testdata/tweets.json.incomplete.gz',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = it.next()
    assert_first_comm(comm)

    (comm, _) = it.next()
    assert_second_comm(comm)

    try:
        it.next()
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_log_config(log_conf, output_file):
    (log_conf_path, log_path) = log_conf
    p = Popen([
        'scripts/tweets2concrete.py',
        '--log-conf-path', log_conf_path,
        '--log-interval', '1',
        'tests/testdata/tweets.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0
    assert len(stdout) == 0
    assert len(stderr) == 0

    with open(log_path) as f:
        data = f.read()
        assert len([line
                    for line in data.strip().split('\n')
                    if 'INFO' in line]) >= 2

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = it.next()
    assert_first_comm(comm)

    (comm, _) = it.next()
    assert_second_comm(comm)

    try:
        it.next()
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_deleted(output_file):
    p = Popen([
        'scripts/tweets2concrete.py',
        'tests/testdata/tweets.deleted.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = it.next()
    assert_first_comm(comm)

    (comm, _) = it.next()
    assert_second_comm(comm)

    try:
        it.next()
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_bad_line(output_file):
    p = Popen([
        'scripts/tweets2concrete.py',
        '--skip-bad-lines',
        'tests/testdata/tweets.bad-line.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = it.next()
    assert_first_comm(comm)

    (comm, _) = it.next()
    assert_second_comm(comm)

    try:
        it.next()
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_bad_line_unicode(output_file):
    p = Popen([
        'scripts/tweets2concrete.py',
        '--skip-bad-lines',
        'tests/testdata/tweets.bad-line-unicode.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = it.next()
    assert_first_comm(comm)

    (comm, _) = it.next()
    assert_second_comm(comm)

    try:
        it.next()
    except StopIteration:
        pass
    else:
        assert False


def test_tweets2concrete_invalid(output_file):
    p = Popen([
        'scripts/tweets2concrete.py',
        '--skip-invalid-comms',
        'tests/testdata/tweets.invalid.json',
        output_file
    ], stdout=PIPE, stderr=PIPE)
    (stdout, stderr) = p.communicate()
    assert p.returncode == 0

    reader = CommunicationReader(output_file)
    it = iter(reader)

    (comm, _) = it.next()
    assert_first_comm(comm)

    (comm, _) = it.next()
    assert_second_comm(comm)

    try:
        it.next()
    except StopIteration:
        pass
    else:
        assert False



import time
import unittest

from testfixtures import LogCapture, StringComparison

import concrete
from concrete.util import add_references_to_communication
from concrete.validate import (
    validate_communication, validate_entity_mention_ids,
    validate_entity_mention_tokenization_ids,
    validate_token_offsets_for_section,
    validate_thrift_deep,
    validate_token_offsets_for_sentence
)
from test_helper import read_test_comm


class TestAddReferences(unittest.TestCase):

    def test_add_references(self):
        comm = read_test_comm()
        add_references_to_communication(comm)


class TestCommunication(unittest.TestCase):

    def test_entity_mention_ids(self):
        comm = read_test_comm()
        self.assertTrue(validate_communication(comm))
        self.assertTrue(validate_entity_mention_ids(comm))

        comm.entitySetList[0].entityList[0].mentionIdList[
            0] = concrete.UUID(uuidString='BAD_ENTITY_MENTION_UUID')

        with LogCapture() as log_capture:
            self.assertFalse(validate_entity_mention_ids(comm))
        log_capture.check(('root', 'ERROR', StringComparison(
            r'.*invalid entityMentionId.*BAD_ENTITY_MENTION_UUID')))

    def test_entity_mention_tokenization(self):
        comm = read_test_comm()
        self.assertTrue(validate_communication(comm))
        self.assertTrue(validate_entity_mention_ids(comm))

        comm.entityMentionSetList[0].mentionList[0].tokens.tokenizationId = (
            concrete.UUID(uuidString='BAD_TOKENIZATION_UUID')
        )

        with LogCapture() as log_capture:
            self.assertFalse(validate_entity_mention_tokenization_ids(comm))
        log_capture.check(('root', 'ERROR', StringComparison(
            r'.*invalid tokenizationId.*BAD_TOKENIZATION_UUID')))


class TestRequiredThriftFields(unittest.TestCase):

    def test_check_required_fields(self):
        # When a field is marked as required in a .thrift file, the
        # Python code generated by the Thrift compiler only seems to
        # capture this requirement in the validate() function for the
        # generated class.  While the ThriftGeneratedClass.thrift_spec
        # structure captures the names and types of the fields,
        # thrift_spec does not seem to store any flags indicating
        # whether or not a field is required.
        #
        # Here is the validate() function for the Communication class:
        #
        #    def validate(self):
        #        if self.id is None:
        #            raise TProtocol.TProtocolException(
        #                message='Required field id is unset!')
        #        if self.uuid is None:
        #            raise TProtocol.TProtocolException(
        #                message='Required field uuid is unset!')
        #        if self.type is None:
        #            raise TProtocol.TProtocolException(
        #                message='Required field type is unset!')
        #        return
        #
        # The validate() function raises an exception when it can't
        # find a required field.  There doesn't seem to be any way to
        # determine whether multiple required fields are missing,
        # aside from assigning a value to the required field and
        # running validate() again.

        comm = concrete.Communication()

        with LogCapture() as log_capture:
            self.assertFalse(
                validate_thrift_deep(comm))
        log_capture.check(
            ('root', 'ERROR',
             "Communication: Required Field 'id' is unset!"))

        comm.id = "ID"
        with LogCapture() as log_capture:
            self.assertFalse(
                validate_thrift_deep(comm))
        log_capture.check(
            ('root', 'ERROR',
             "Communication: Required Field 'uuid' is unset!"))

        comm.uuid = concrete.UUID(uuidString="TEST_UUID")
        with LogCapture() as log_capture:
            self.assertFalse(
                validate_thrift_deep(comm))
        log_capture.check(('root', 'ERROR', StringComparison(
            r".*TEST_UUID.*Required Field 'type' is unset!")))

        comm.metadata = concrete.AnnotationMetadata(
            tool="TEST", timestamp=int(time.time()))

        comm.type = "OTHER"
        self.assertTrue(
            validate_thrift_deep(comm))


class TestTextspanOffsets(unittest.TestCase):

    def test_validate_token_offsets_for_good_sentence(self):
        sentence = self.create_sentence_with_token(0, 30, 0, 10)
        self.assertTrue(validate_token_offsets_for_sentence(sentence))

    def test_validate_token_offsets_for_sentence_rev_sentence_offsets(self):
        sentence = self.create_sentence_with_token(30, 0, 0, 10)
        with LogCapture():
            self.assertFalse(validate_token_offsets_for_sentence(sentence))

    def test_validate_token_offsets_for_sentence_rev_token_offsets(self):
        sentence = self.create_sentence_with_token(0, 30, 10, 0)
        with LogCapture():
            self.assertFalse(validate_token_offsets_for_sentence(sentence))

    def test_validate_token_offsets_for_sentence_token_outside(self):
        sentence = self.create_sentence_with_token(0, 30, 25, 35)
        with LogCapture():
            self.assertFalse(validate_token_offsets_for_sentence(sentence))

    def test_validate_token_offsets_for_good_section(self):
        section = self.create_section_with_sentence(0, 30, 0, 10)
        self.assertTrue(validate_token_offsets_for_section(section))

    def test_validate_token_offsets_for_section_rev_section_offsets(self):
        section = self.create_section_with_sentence(30, 0, 0, 10)
        with LogCapture():
            self.assertFalse(validate_token_offsets_for_section(section))

    def test_validate_token_offsets_for_section_rev_token_offsets(self):
        section = self.create_section_with_sentence(0, 30, 10, 0)
        with LogCapture():
            self.assertFalse(validate_token_offsets_for_section(section))

    def test_validate_token_offsets_for_section_sentence_outside(self):
        section = self.create_section_with_sentence(0, 30, 25, 35)
        with LogCapture():
            self.assertFalse(validate_token_offsets_for_section(section))

    def test_validate_token_offsets_for_real_example_section_data(self):
        section = self.create_section_with_sentence(55, 296, 0, 118)
        with LogCapture():
            self.assertFalse(validate_token_offsets_for_section(section))

    def create_section_with_sentence(self, section_start, section_ending,
                                     sentence_start, sentence_ending):
        sentence_textspan = concrete.spans.ttypes.TextSpan(
            start=sentence_start, ending=sentence_ending)
        sentence = concrete.structure.ttypes.Sentence(
            textSpan=sentence_textspan, uuid='TEST_SENTENCE')
        section_textspan = concrete.spans.ttypes.TextSpan(
            start=section_start, ending=section_ending)
        section = concrete.structure.ttypes.Section(
            sentenceList=[sentence],
            textSpan=section_textspan,
            uuid='TEST_SECTION')
        return section

    def create_sentence_with_token(self, sentence_start, sentence_ending,
                                   token_start, token_ending):
        token_textspan = concrete.spans.ttypes.TextSpan(
            start=token_start, ending=token_ending)
        token = concrete.structure.ttypes.Token(textSpan=token_textspan)
        tokenization = concrete.structure.ttypes.Tokenization(
            tokenList=concrete.TokenList(tokenList=[token]))
        sentence_textspan = concrete.spans.ttypes.TextSpan(
            start=sentence_start, ending=sentence_ending)
        sentence = concrete.structure.ttypes.Sentence(
            tokenization=tokenization,
            textSpan=sentence_textspan,
            uuid='TEST')
        return sentence



# This file is NOT auto-generated

"""
Python modules and scripts for working with
[Concrete](https://github.com/hltcoe/concrete), an HLT data
specification defined using [Thrift](http://thrift.apache.org).
"""


from audio.ttypes import *  # noqa
from clustering.ttypes import *  # noqa
from communication.ttypes import *  # noqa
from email.ttypes import *  # noqa
from entities.ttypes import *  # noqa
from exceptions.ttypes import *  # noqa
from language.ttypes import *  # noqa
from linking.ttypes import *  # noqa
from metadata.ttypes import *  # noqa
from nitf.ttypes import *  # noqa
from services.ttypes import *  # noqa
from situations.ttypes import *  # noqa
from spans.ttypes import *  # noqa
from structure.ttypes import *  # noqa
from twitter.ttypes import *  # noqa
from uuid.ttypes import *  # noqa

from version import __version__  # noqa



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.9.3)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:new_style,utf8strings,coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException
import logging
from ttypes import *
from thrift.Thrift import TProcessor
from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
  from thrift.protocol import fastbinary
except:
  fastbinary = None


class Iface(object):
  """
  DEPRECATION NOTICE: This interface has been replaced by
  the ServiceAnnotation interface in the concrete-services
  package. All new development based on thrift services
  and concrete should be based on that repo. This is only
  left for historical, yet-to-be-updated services: do not
  implement new services with this interface.

  Annotator service methods. For concrete analytics that
  are to be stood up as independent services, accessible
  from any programming language.
  """
  def annotate(self, original):
    """
    Main annotation method. Takes a communication as input
    and returns a new one as output.

    It is up to the implementing service to verify that
    the input communication is valid.

    Can throw a ConcreteThriftException upon error
    (invalid input, analytic exception, etc.).

    Parameters:
     - original
    """
    pass

  def getMetadata(self):
    """
    Return the tool's AnnotationMetadata.
    """
    pass

  def getDocumentation(self):
    """
    Return a detailed description of what the particular tool
    does, what inputs and outputs to expect, etc.

    Developers whom are not familiar with the particular
    analytic should be able to read this string and
    understand the essential functions of the analytic.
    """
    pass

  def shutdown(self):
    """
    Indicate to the server it should shut down.
    """
    pass


class Client(Iface):
  """
  DEPRECATION NOTICE: This interface has been replaced by
  the ServiceAnnotation interface in the concrete-services
  package. All new development based on thrift services
  and concrete should be based on that repo. This is only
  left for historical, yet-to-be-updated services: do not
  implement new services with this interface.

  Annotator service methods. For concrete analytics that
  are to be stood up as independent services, accessible
  from any programming language.
  """
  def __init__(self, iprot, oprot=None):
    self._iprot = self._oprot = iprot
    if oprot is not None:
      self._oprot = oprot
    self._seqid = 0

  def annotate(self, original):
    """
    Main annotation method. Takes a communication as input
    and returns a new one as output.

    It is up to the implementing service to verify that
    the input communication is valid.

    Can throw a ConcreteThriftException upon error
    (invalid input, analytic exception, etc.).

    Parameters:
     - original
    """
    self.send_annotate(original)
    return self.recv_annotate()

  def send_annotate(self, original):
    self._oprot.writeMessageBegin('annotate', TMessageType.CALL, self._seqid)
    args = annotate_args()
    args.original = original
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_annotate(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = annotate_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    if result.ex is not None:
      raise result.ex
    raise TApplicationException(TApplicationException.MISSING_RESULT, "annotate failed: unknown result")

  def getMetadata(self):
    """
    Return the tool's AnnotationMetadata.
    """
    self.send_getMetadata()
    return self.recv_getMetadata()

  def send_getMetadata(self):
    self._oprot.writeMessageBegin('getMetadata', TMessageType.CALL, self._seqid)
    args = getMetadata_args()
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_getMetadata(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = getMetadata_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "getMetadata failed: unknown result")

  def getDocumentation(self):
    """
    Return a detailed description of what the particular tool
    does, what inputs and outputs to expect, etc.

    Developers whom are not familiar with the particular
    analytic should be able to read this string and
    understand the essential functions of the analytic.
    """
    self.send_getDocumentation()
    return self.recv_getDocumentation()

  def send_getDocumentation(self):
    self._oprot.writeMessageBegin('getDocumentation', TMessageType.CALL, self._seqid)
    args = getDocumentation_args()
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

  def recv_getDocumentation(self):
    iprot = self._iprot
    (fname, mtype, rseqid) = iprot.readMessageBegin()
    if mtype == TMessageType.EXCEPTION:
      x = TApplicationException()
      x.read(iprot)
      iprot.readMessageEnd()
      raise x
    result = getDocumentation_result()
    result.read(iprot)
    iprot.readMessageEnd()
    if result.success is not None:
      return result.success
    raise TApplicationException(TApplicationException.MISSING_RESULT, "getDocumentation failed: unknown result")

  def shutdown(self):
    """
    Indicate to the server it should shut down.
    """
    self.send_shutdown()

  def send_shutdown(self):
    self._oprot.writeMessageBegin('shutdown', TMessageType.ONEWAY, self._seqid)
    args = shutdown_args()
    args.write(self._oprot)
    self._oprot.writeMessageEnd()
    self._oprot.trans.flush()

class Processor(Iface, TProcessor):
  def __init__(self, handler):
    self._handler = handler
    self._processMap = {}
    self._processMap["annotate"] = Processor.process_annotate
    self._processMap["getMetadata"] = Processor.process_getMetadata
    self._processMap["getDocumentation"] = Processor.process_getDocumentation
    self._processMap["shutdown"] = Processor.process_shutdown

  def process(self, iprot, oprot):
    (name, type, seqid) = iprot.readMessageBegin()
    if name not in self._processMap:
      iprot.skip(TType.STRUCT)
      iprot.readMessageEnd()
      x = TApplicationException(TApplicationException.UNKNOWN_METHOD, 'Unknown function %s' % (name))
      oprot.writeMessageBegin(name, TMessageType.EXCEPTION, seqid)
      x.write(oprot)
      oprot.writeMessageEnd()
      oprot.trans.flush()
      return
    else:
      self._processMap[name](self, seqid, iprot, oprot)
    return True

  def process_annotate(self, seqid, iprot, oprot):
    args = annotate_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = annotate_result()
    try:
      result.success = self._handler.annotate(args.original)
      msg_type = TMessageType.REPLY
    except (TTransport.TTransportException, KeyboardInterrupt, SystemExit):
      raise
    except concrete.exceptions.ttypes.ConcreteThriftException as ex:
      msg_type = TMessageType.REPLY
      result.ex = ex
    except Exception as ex:
      msg_type = TMessageType.EXCEPTION
      logging.exception(ex)
      result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
    oprot.writeMessageBegin("annotate", msg_type, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_getMetadata(self, seqid, iprot, oprot):
    args = getMetadata_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = getMetadata_result()
    try:
      result.success = self._handler.getMetadata()
      msg_type = TMessageType.REPLY
    except (TTransport.TTransportException, KeyboardInterrupt, SystemExit):
      raise
    except Exception as ex:
      msg_type = TMessageType.EXCEPTION
      logging.exception(ex)
      result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
    oprot.writeMessageBegin("getMetadata", msg_type, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_getDocumentation(self, seqid, iprot, oprot):
    args = getDocumentation_args()
    args.read(iprot)
    iprot.readMessageEnd()
    result = getDocumentation_result()
    try:
      result.success = self._handler.getDocumentation()
      msg_type = TMessageType.REPLY
    except (TTransport.TTransportException, KeyboardInterrupt, SystemExit):
      raise
    except Exception as ex:
      msg_type = TMessageType.EXCEPTION
      logging.exception(ex)
      result = TApplicationException(TApplicationException.INTERNAL_ERROR, 'Internal error')
    oprot.writeMessageBegin("getDocumentation", msg_type, seqid)
    result.write(oprot)
    oprot.writeMessageEnd()
    oprot.trans.flush()

  def process_shutdown(self, seqid, iprot, oprot):
    args = shutdown_args()
    args.read(iprot)
    iprot.readMessageEnd()
    try:
      self._handler.shutdown()
      msg_type = TMessageType.REPLY
    except (TTransport.TTransportException, KeyboardInterrupt, SystemExit):
      raise
    except:
      pass


# HELPER FUNCTIONS AND STRUCTURES

class annotate_args(object):
  """
  Attributes:
   - original
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'original', (concrete.communication.ttypes.Communication, concrete.communication.ttypes.Communication.thrift_spec), None, ), # 1
  )

  def __init__(self, original=None,):
    self.original = original

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.original = concrete.communication.ttypes.Communication()
          self.original.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('annotate_args')
    if self.original is not None:
      oprot.writeFieldBegin('original', TType.STRUCT, 1)
      self.original.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.original)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class annotate_result(object):
  """
  Attributes:
   - success
   - ex
  """

  thrift_spec = (
    (0, TType.STRUCT, 'success', (concrete.communication.ttypes.Communication, concrete.communication.ttypes.Communication.thrift_spec), None, ), # 0
    (1, TType.STRUCT, 'ex', (concrete.exceptions.ttypes.ConcreteThriftException, concrete.exceptions.ttypes.ConcreteThriftException.thrift_spec), None, ), # 1
  )

  def __init__(self, success=None, ex=None,):
    self.success = success
    self.ex = ex

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.STRUCT:
          self.success = concrete.communication.ttypes.Communication()
          self.success.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 1:
        if ftype == TType.STRUCT:
          self.ex = concrete.exceptions.ttypes.ConcreteThriftException()
          self.ex.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('annotate_result')
    if self.success is not None:
      oprot.writeFieldBegin('success', TType.STRUCT, 0)
      self.success.write(oprot)
      oprot.writeFieldEnd()
    if self.ex is not None:
      oprot.writeFieldBegin('ex', TType.STRUCT, 1)
      self.ex.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    value = (value * 31) ^ hash(self.ex)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class getMetadata_args(object):

  thrift_spec = (
  )

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('getMetadata_args')
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class getMetadata_result(object):
  """
  Attributes:
   - success
  """

  thrift_spec = (
    (0, TType.STRUCT, 'success', (concrete.metadata.ttypes.AnnotationMetadata, concrete.metadata.ttypes.AnnotationMetadata.thrift_spec), None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.STRUCT:
          self.success = concrete.metadata.ttypes.AnnotationMetadata()
          self.success.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('getMetadata_result')
    if self.success is not None:
      oprot.writeFieldBegin('success', TType.STRUCT, 0)
      self.success.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class getDocumentation_args(object):

  thrift_spec = (
  )

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('getDocumentation_args')
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class getDocumentation_result(object):
  """
  Attributes:
   - success
  """

  thrift_spec = (
    (0, TType.STRING, 'success', None, None, ), # 0
  )

  def __init__(self, success=None,):
    self.success = success

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 0:
        if ftype == TType.STRING:
          self.success = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('getDocumentation_result')
    if self.success is not None:
      oprot.writeFieldBegin('success', TType.STRING, 0)
      oprot.writeString(self.success.encode('utf-8'))
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.success)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class shutdown_args(object):

  thrift_spec = (
  )

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('shutdown_args')
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    return


  def __hash__(self):
    value = 17
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)



from multiprocessing import Process
from time import sleep
from socket import create_connection

from concrete.services import Annotator
from concrete.util.thrift_factory import factory


class AnnotatorClientWrapper(object):
    """
    A sample client implementation of the Concrete Annotator service.

    Provides sensible/current defaults for transport and protocol.
    """

    def __init__(self, host, port):
        self.host = host
        self.port = port

    def __enter__(self):
        socket = factory.createSocket(self.host, self.port)
        self.transport = factory.createTransport(socket)
        protocol = factory.createProtocol(self.transport)

        cli = Annotator.Client(protocol)

        self.transport.open()
        return cli

    def __exit__(self, type, value, traceback):
        self.transport.close()


class AnnotatorServiceWrapper(object):
    """
    A sample wrapper around the Concrete annotator
    service, providing an easy way to wrap an implementation
    of the Annotator service.
    """

    def __init__(self, implementation):
        self.processor = Annotator.Processor(implementation)

    def serve(self, host, port):
        server = factory.createServer(self.processor, host, port)

        # NOTE: Thrift's servers run indefinitely. This server implementation
        # may be killed by a KeyboardInterrupt (Control-C); otherwise, the
        # process must be killed to terminate the server.
        server.serve()


class SubprocessAnnotatorServiceWrapper(object):
    '''
    Annotator service wrapper that runs server in a subprocess via a
    context manager interface.
    '''

    SLEEP_INTERVAL = 0.1

    def __init__(self, implementation, host, port, timeout=None):
        self.proc = None
        self.server = AnnotatorServiceWrapper(implementation)
        self.host = host
        self.port = port
        self.timeout = timeout

    def __enter__(self):
        self.proc = Process(target=self.server.serve,
                            args=(self.host, self.port))
        self.proc.start()

        s = None
        elapsed = 0.

        # Loop until successful connect or timeout
        while s is None and (self.timeout is None or elapsed < self.timeout):
            try:
                s = create_connection((self.host, self.port))
                s.close()
            except:
                s = None

                # this is not precise... not important
                sleep(self.SLEEP_INTERVAL)
                elapsed += self.SLEEP_INTERVAL

        return self

    def __exit__(self, type, value, traceback):
        self.proc.terminate()
        self.proc.join()



__version__ = '4.10.6'


def concrete_library_version():
    return __version__


def concrete_schema_version():
    return ".".join(__version__.split(".")[0:2])


def add_argparse_argument(parser):
    parser.add_argument("--version", action="version",
                        version=("Concrete schema version: %s, "
                                 "Concrete python library version: %s") %
                                (concrete_schema_version(),
                                 concrete_library_version()))



# -*- coding: utf-8 -*-
#
# Autogenerated by Thrift Compiler (0.9.3)
#
# DO NOT EDIT UNLESS YOU ARE SURE THAT YOU KNOW WHAT YOU ARE DOING
#
#  options string: py:new_style,utf8strings,coding=utf-8
#

from thrift.Thrift import TType, TMessageType, TException, TApplicationException
import concrete.services.ttypes
import concrete.uuid.ttypes
import concrete.communication.ttypes


from thrift.transport import TTransport
from thrift.protocol import TBinaryProtocol, TProtocol
try:
  from thrift.protocol import fastbinary
except:
  fastbinary = None



class AnnotationTask(object):
  """
  Annotation task including information for pulling data.

  Attributes:
   - type: Type of annotation task
   - language: Language of the data for the task
   - unitType: Entire communication or individual sentences
   - units: Identifiers for each annotation unit
  """

  thrift_spec = (
    None, # 0
    (1, TType.I32, 'type', None, None, ), # 1
    (2, TType.STRING, 'language', None, None, ), # 2
    (3, TType.I32, 'unitType', None, None, ), # 3
    (4, TType.LIST, 'units', (TType.STRUCT,(concrete.services.ttypes.AnnotationUnitIdentifier, concrete.services.ttypes.AnnotationUnitIdentifier.thrift_spec)), None, ), # 4
  )

  def __init__(self, type=None, language=None, unitType=None, units=None,):
    self.type = type
    self.language = language
    self.unitType = unitType
    self.units = units

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.I32:
          self.type = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRING:
          self.language = iprot.readString().decode('utf-8')
        else:
          iprot.skip(ftype)
      elif fid == 3:
        if ftype == TType.I32:
          self.unitType = iprot.readI32()
        else:
          iprot.skip(ftype)
      elif fid == 4:
        if ftype == TType.LIST:
          self.units = []
          (_etype3, _size0) = iprot.readListBegin()
          for _i4 in xrange(_size0):
            _elem5 = concrete.services.ttypes.AnnotationUnitIdentifier()
            _elem5.read(iprot)
            self.units.append(_elem5)
          iprot.readListEnd()
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('AnnotationTask')
    if self.type is not None:
      oprot.writeFieldBegin('type', TType.I32, 1)
      oprot.writeI32(self.type)
      oprot.writeFieldEnd()
    if self.language is not None:
      oprot.writeFieldBegin('language', TType.STRING, 2)
      oprot.writeString(self.language.encode('utf-8'))
      oprot.writeFieldEnd()
    if self.unitType is not None:
      oprot.writeFieldBegin('unitType', TType.I32, 3)
      oprot.writeI32(self.unitType)
      oprot.writeFieldEnd()
    if self.units is not None:
      oprot.writeFieldBegin('units', TType.LIST, 4)
      oprot.writeListBegin(TType.STRUCT, len(self.units))
      for iter6 in self.units:
        iter6.write(oprot)
      oprot.writeListEnd()
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.type is None:
      raise TProtocol.TProtocolException(message='Required field type is unset!')
    if self.unitType is None:
      raise TProtocol.TProtocolException(message='Required field unitType is unset!')
    if self.units is None:
      raise TProtocol.TProtocolException(message='Required field units is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.type)
    value = (value * 31) ^ hash(self.language)
    value = (value * 31) ^ hash(self.unitType)
    value = (value * 31) ^ hash(self.units)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)

class Annotation(object):
  """
  Annotation on a communication.

  Attributes:
   - id: Identifier of the part of the communication being annotated.
   - communication: Communication with the annotation stored in it.
  The location of the annotation depends on the annotation unit identifier
  """

  thrift_spec = (
    None, # 0
    (1, TType.STRUCT, 'id', (concrete.services.ttypes.AnnotationUnitIdentifier, concrete.services.ttypes.AnnotationUnitIdentifier.thrift_spec), None, ), # 1
    (2, TType.STRUCT, 'communication', (concrete.communication.ttypes.Communication, concrete.communication.ttypes.Communication.thrift_spec), None, ), # 2
  )

  def __init__(self, id=None, communication=None,):
    self.id = id
    self.communication = communication

  def read(self, iprot):
    if iprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and isinstance(iprot.trans, TTransport.CReadableTransport) and self.thrift_spec is not None and fastbinary is not None:
      fastbinary.decode_binary(self, iprot.trans, (self.__class__, self.thrift_spec))
      return
    iprot.readStructBegin()
    while True:
      (fname, ftype, fid) = iprot.readFieldBegin()
      if ftype == TType.STOP:
        break
      if fid == 1:
        if ftype == TType.STRUCT:
          self.id = concrete.services.ttypes.AnnotationUnitIdentifier()
          self.id.read(iprot)
        else:
          iprot.skip(ftype)
      elif fid == 2:
        if ftype == TType.STRUCT:
          self.communication = concrete.communication.ttypes.Communication()
          self.communication.read(iprot)
        else:
          iprot.skip(ftype)
      else:
        iprot.skip(ftype)
      iprot.readFieldEnd()
    iprot.readStructEnd()

  def write(self, oprot):
    if oprot.__class__ == TBinaryProtocol.TBinaryProtocolAccelerated and self.thrift_spec is not None and fastbinary is not None:
      oprot.trans.write(fastbinary.encode_binary(self, (self.__class__, self.thrift_spec)))
      return
    oprot.writeStructBegin('Annotation')
    if self.id is not None:
      oprot.writeFieldBegin('id', TType.STRUCT, 1)
      self.id.write(oprot)
      oprot.writeFieldEnd()
    if self.communication is not None:
      oprot.writeFieldBegin('communication', TType.STRUCT, 2)
      self.communication.write(oprot)
      oprot.writeFieldEnd()
    oprot.writeFieldStop()
    oprot.writeStructEnd()

  def validate(self):
    if self.id is None:
      raise TProtocol.TProtocolException(message='Required field id is unset!')
    if self.communication is None:
      raise TProtocol.TProtocolException(message='Required field communication is unset!')
    return


  def __hash__(self):
    value = 17
    value = (value * 31) ^ hash(self.id)
    value = (value * 31) ^ hash(self.communication)
    return value

  def __repr__(self):
    L = ['%s=%r' % (key, value)
      for key, value in self.__dict__.iteritems()]
    return '%s(%s)' % (self.__class__.__name__, ', '.join(L))

  def __eq__(self, other):
    return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

  def __ne__(self, other):
    return not (self == other)



import unittest

from concrete.structure.ttypes import (
    TokenizationKind, TokenLattice, LatticePath, Token,
    TokenTagging, TaggedToken, Tokenization
)

from concrete.util.simple_comm import create_comm
from concrete.util.tokenization import (
    get_tokens, get_pos, get_lemmas, get_tagged_tokens
)


class TestTokenization(unittest.TestCase):
    """
    Test suite for module.
    """

    def test_get_tokens_invalid_kind(self):
        self.assertRaises(ValueError,
                          get_tokens,
                          Tokenization(kind='invalid-kind'))


class TestGetTaggedTokens(unittest.TestCase):
    def setUp(self):
        self.tokenization = Tokenization(
            tokenTaggingList=[
                TokenTagging(
                    taggingType='?',
                    taggedTokenList=[
                        TaggedToken(tokenIndex=0, tag='?'),
                        TaggedToken(tokenIndex=1, tag='?'),
                        TaggedToken(tokenIndex=2, tag='?'),
                    ],
                ),
                TokenTagging(
                    taggingType='POS',
                    taggedTokenList=[
                        TaggedToken(tokenIndex=0, tag='N'),
                        TaggedToken(tokenIndex=1, tag='N'),
                        TaggedToken(tokenIndex=2, tag='X'),
                    ],
                ),
                TokenTagging(
                    taggingType='NUMERAL',
                    taggedTokenList=[
                        TaggedToken(tokenIndex=0, tag='N'),
                        TaggedToken(tokenIndex=1, tag='N'),
                        TaggedToken(tokenIndex=2, tag='Y'),
                    ],
                ),
                TokenTagging(
                    taggingType='LEMMA',
                    taggedTokenList=[
                        TaggedToken(tokenIndex=0, tag='mambo'),
                        TaggedToken(tokenIndex=1, tag='number'),
                        TaggedToken(tokenIndex=2, tag='4'),
                    ],
                ),
            ],
        )

    def test_get_pos(self):
        self.assertEqual(
            ['N', 'N', 'X'],
            map(lambda t: t.tag, get_pos(self.tokenization))
        )
        self.assertEqual(
            [0, 1, 2],
            map(lambda t: t.tokenIndex, get_pos(self.tokenization))
        )

    def test_get_pos_no_tagging(self):
        self.tokenization.tokenTaggingList = filter(
            lambda ttl: ttl.taggingType != 'POS',
            self.tokenization.tokenTaggingList
        )
        with self.assertRaises(Exception):
            get_pos(self.tokenization)

    def test_get_pos_non_unique_tagging(self):
        self.tokenization.tokenTaggingList.append(
            TokenTagging(
                taggingType='POS',
                taggedTokenList=[
                    TaggedToken(tokenIndex=0, tag='N'),
                    TaggedToken(tokenIndex=1, tag='X'),
                    TaggedToken(tokenIndex=2, tag='N'),
                ],
            ),
        )
        with self.assertRaises(Exception):
            get_pos(self.tokenization)

    def test_get_lemmas(self):
        self.assertEqual(
            ['mambo', 'number', '4'],
            map(lambda t: t.tag, get_lemmas(self.tokenization))
        )
        self.assertEqual(
            [0, 1, 2],
            map(lambda t: t.tokenIndex, get_lemmas(self.tokenization))
        )

    def test_get_lemmas_no_tagging(self):
        self.tokenization.tokenTaggingList = filter(
            lambda ttl: ttl.taggingType != 'LEMMA',
            self.tokenization.tokenTaggingList
        )
        with self.assertRaises(Exception):
            get_lemmas(self.tokenization)

    def test_get_lemmas_non_unique_tagging(self):
        self.tokenization.tokenTaggingList.append(
            TokenTagging(
                taggingType='LEMMA',
                taggedTokenList=[
                    TaggedToken(tokenIndex=0, tag='mambo'),
                    TaggedToken(tokenIndex=1, tag='number'),
                    TaggedToken(tokenIndex=2, tag='four'),
                ],
            ),
        )
        with self.assertRaises(Exception):
            get_lemmas(self.tokenization)

    def test_get_tagged_tokens(self):
        self.assertEqual(
            ['N', 'N', 'Y'],
            map(lambda t: t.tag, get_tagged_tokens(self.tokenization,
                                                   'NUMERAL'))
        )
        self.assertEqual(
            [0, 1, 2],
            map(lambda t: t.tokenIndex, get_tagged_tokens(self.tokenization,
                                                          'NUMERAL'))
        )

    def test_get_tagged_tokens_no_tagging(self):
        self.tokenization.tokenTaggingList = filter(
            lambda ttl: ttl.taggingType != 'NUMERAL',
            self.tokenization.tokenTaggingList
        )
        with self.assertRaises(Exception):
            get_tagged_tokens(self.tokenization, 'NUMERAL')

    def test_get_tagged_tokens_non_unique_tagging(self):
        self.tokenization.tokenTaggingList.append(
            TokenTagging(
                taggingType='NUMERAL',
                taggedTokenList=[
                    TaggedToken(tokenIndex=0, tag='N'),
                    TaggedToken(tokenIndex=1, tag='Y'),
                    TaggedToken(tokenIndex=2, tag='Y'),
                ],
            ),
        )
        with self.assertRaises(Exception):
            get_tagged_tokens(self.tokenization, 'NUMERAL')


class TestGetTokenList(unittest.TestCase):
    def test_no_lattice(self):
        comm = create_comm('comm-1', 'mambo no. 4')
        tokenization = comm.sectionList[0].sentenceList[0].tokenization
        tokenization.kind = None
        token_texts = [t.text for t in get_tokens(tokenization)]
        self.assertEqual(['mambo', 'no.', '4'], token_texts)

    def test_no_lattice_with_no_kind(self):
        comm = create_comm('comm-1', 'mambo no. 4')
        tokenization = comm.sectionList[0].sentenceList[0].tokenization
        token_texts = [t.text for t in get_tokens(tokenization)]
        self.assertEqual(['mambo', 'no.', '4'], token_texts)

    def test_lattice_with_token_list_kind(self):
        comm = create_comm('comm-1', 'mambo no. 4')
        tokenization = comm.sectionList[0].sentenceList[0].tokenization
        lattice_path = LatticePath()
        lattice_path.tokenList = [Token(tokenIndex=0, text='mambo'),
                                  Token(tokenIndex=0, text='no.'),
                                  Token(tokenIndex=0, text='3')]
        token_lattice = TokenLattice()
        token_lattice.cachedBestPath = lattice_path
        tokenization.lattice = token_lattice
        token_texts = [t.text for t in get_tokens(tokenization)]
        self.assertEqual(['mambo', 'no.', '4'], token_texts)

    def test_lattice_with_token_lattice_kind(self):
        comm = create_comm('comm-1', 'mambo no. 4')
        tokenization = comm.sectionList[0].sentenceList[0].tokenization
        lattice_path = LatticePath()
        lattice_path.tokenList = [Token(tokenIndex=0, text='mambo'),
                                  Token(tokenIndex=0, text='no.'),
                                  Token(tokenIndex=0, text='3')]
        token_lattice = TokenLattice()
        token_lattice.cachedBestPath = lattice_path
        tokenization.lattice = token_lattice
        tokenization.kind = TokenizationKind.TOKEN_LATTICE
        token_texts = [t.text for t in get_tokens(tokenization)]
        self.assertEqual(['mambo', 'no.', '3'], token_texts)

    def test_lattice_with_no_kind(self):
        comm = create_comm('comm-1', 'mambo no. 4')
        tokenization = comm.sectionList[0].sentenceList[0].tokenization
        lattice_path = LatticePath()
        lattice_path.tokenList = [Token(tokenIndex=0, text='mambo'),
                                  Token(tokenIndex=0, text='no.'),
                                  Token(tokenIndex=0, text='3')]
        token_lattice = TokenLattice()
        token_lattice.cachedBestPath = lattice_path
        tokenization.lattice = token_lattice
        tokenization.kind = None
        token_texts = [t.text for t in get_tokens(tokenization)]
        self.assertEqual(['mambo', 'no.', '4'], token_texts)



from pytest import fixture

from concrete.util.twitter import (json_tweet_object_to_Communication,
                                   twitter_lid_to_iso639_3)


TWEET_TXT = ("Barber tells me - his son is colorblind / my hair is auburn /"
             " and auburn is a shade of green")

TWEET_ID = u'238426131689242624'


@fixture
def tweet():
    return {u"text": TWEET_TXT,
            u"id_str": TWEET_ID,
            u'created_at': 'Wed Aug 27 13:08:45 +0000 2008',
            u"user": {u"screen_name": "charman",
                      u"lang": "ja",
                      u'verified': False},
            u'entities': {u"symbols": [],
                          u'hashtags': [],
                          u'user_mentions': [],
                          u'urls': []},
            u'coordinates': None,
            u'place': None,
            u"lang": u"en"}


def test_twitter_lid_conversion():
    assert 'eng' == twitter_lid_to_iso639_3('en-gb')
    assert 'msa' == twitter_lid_to_iso639_3('msa')
    assert 'hun' == twitter_lid_to_iso639_3('hu')
    assert 'zho' == twitter_lid_to_iso639_3('zh-cn')
    assert 'zho' == twitter_lid_to_iso639_3('zh-tw')
    assert 'und' == twitter_lid_to_iso639_3('und')
    assert 'eng' == twitter_lid_to_iso639_3('eng')


def test_json_tweet_object_to_Communication_with_lang(tweet):
    comm = json_tweet_object_to_Communication(tweet)
    assert TWEET_ID == comm.id
    assert TWEET_TXT == comm.text
    assert 1 == len(comm.lidList)
    assert 1219842525 == comm.startTime
    assert 1219842525 == comm.endTime
    assert "jpn" == comm.communicationMetadata.tweetInfo.user.lang
    kvm = comm.lidList[0].languageToProbabilityMap
    assert "eng" == kvm.keys()[0]
    assert 1.0 == kvm["eng"]


def test_json_tweet_object_to_Communication_without_lang(tweet):
    del tweet['lang']
    comm = json_tweet_object_to_Communication(tweet)
    assert comm.lidList is None
